use std::ops::{Deref, DerefMut};

use layout::LayoutNodeAccess;
use platform::{
    render::{Painter, Surface},
    EventInterface, RenderInterface,
};

pub mod component;
pub mod geom;
pub mod input;
pub mod layout;
pub mod platform;
pub mod widget;

pub mod prelude {
    pub use crate::component::Component;
    pub use crate::widget::Widget;
}
use prelude::*;

pub enum UIControlMsg {
    Terminate,
}

pub trait UIComponent: Sized + component::Component<ParentMsg = UIControlMsg> {
    fn build(bc: &widget::BuildContext) -> Self;

    fn map_ui_event(&self, uievent: platform::event::UIEvent) -> Option<Self::Msg>;

    type RootWidget: widget::Widget<Self>;
    fn root_widget(&self) -> &Self::RootWidget;
    fn root_widget_mut(&mut self) -> &mut Self::RootWidget;
}

pub struct UI<PT: platform::PlatformTuple, UIC: UIComponent> {
    layout_cache: std::rc::Rc<layout::LayoutCache>,
    root_layout_node: layout::LayoutNode,
    platform: PT,
    ui_component: UIC,
    input_state: input::InputState,
    ui_running: bool,
}

impl<PT: platform::PlatformTuple, UIC: UIComponent> UI<PT, UIC> {
    pub fn new(pt: PT) -> Self {
        let lcache = layout::LayoutCache::new();
        let bc = widget::BuildContext {
            lcache: lcache.clone(),
        };
        Self {
            root_layout_node: layout::LayoutNode::new(lcache.clone()),
            layout_cache: lcache,
            platform: pt,
            ui_component: UIC::build(&bc),
            input_state: Default::default(),
            ui_running: true,
        }
    }

    pub fn is_ui_running(&self) -> bool {
        self.ui_running
    }

    pub fn ui_component(&self) -> &UIC {
        &self.ui_component
    }
    pub fn ui_component_mut(&mut self) -> &mut UIC {
        &mut self.ui_component
    }

    /// Public helper function, delivers a message to the UI component and then processes whatever
    /// control messages result.
    pub fn deliver_msg(&mut self, msg: UIC::Msg) {
        let cmsgs = self.ui_component.process(msg);
        self.process_control_msgs(cmsgs.into_iter());
    }

    fn process_control_msgs(&mut self, msgs: impl Iterator<Item = UIControlMsg>) {
        for cmsg in msgs {
            match cmsg {
                UIControlMsg::Terminate => {
                    self.ui_running = false;
                }
            }
        }
    }

    pub fn update(&mut self) {
        let mut platform_events = vec![];
        while let Some(evt) = self.platform.event_mut().poll_event() {
            match evt {
                platform::event::WindowEvent::CloseRequest => {
                    platform_events.extend(
                        self.ui_component()
                            .map_ui_event(platform::event::UIEvent::CloseRequest)
                            .into_iter(),
                    );
                }
                _ => {}
            }
        }

        // poll for state changes due to user input
        let poll_msgs = self
            .ui_component
            .root_widget_mut()
            .poll(Some(&self.input_state))
            .into_iter();
        let cmsgs = self
            .ui_component
            .process_all(platform_events.into_iter().chain(poll_msgs));
        self.process_control_msgs(cmsgs.into_iter());

        let tgt_size = self.platform.render_mut().render_target_mut().size();
        self.root_layout_node
            .set_behaviour(layout::NodeBehaviour::Fixed {
                rect: geom::IRect::new_from_size(geom::IPoint::ORIGIN, tgt_size),
            });

        // first, make sure we can paint to the render target
        let (ri, ti) = self.platform.render_text_mut();
        if let Some(mut painter) = ri.render_target_mut().painter() {
            // perform layout as needed
            layout::recalculate(LayoutNodeAccess::new(&layout::LinearAccess::new(
                &self.root_layout_node,
                self.ui_component.root_widget().layout_node(),
            )));

            // now render as needed
            self.ui_component.root_widget().render(painter.as_mut(), ti);
        } else {
            log::warn!("Could not draw UI: no painter available");
        }
    }

    pub fn wait(&mut self) {
        self.platform.event_mut().wait_for_event();
    }

    pub fn run(mut self) {
        while self.ui_running {
            self.update();
            self.wait();
        }
    }
}
