use std::collections::HashMap;

use super::{Widget, WidgetID};

#[derive(Default)]
pub struct WidgetContext<Message> {
    mappers: HashMap<WidgetID, Box<dyn std::any::Any>>,
    queue: Vec<Message>,
    _ghost: std::marker::PhantomData<Message>,
}

impl<Message: 'static> WidgetContext<Message> {
    pub fn register<W: Widget>(&mut self, id: WidgetID, cb: impl Fn(W::Event) -> Option<Message> + 'static) {
        let mapper = Box::new(cb) as Box<dyn Fn(W::Event) -> Option<Message>>;
        self.mappers.insert(id, Box::new(mapper));
    }
    pub fn clear(&mut self, id: WidgetID) {
        self.mappers.remove(&id);
    }
}

pub trait WidgetContextAccess<Event> {
    fn sink(&mut self, id: WidgetID, event: Event);
}

impl<Event: 'static, Message: 'static> WidgetContextAccess<Event> for WidgetContext<Message> {
    fn sink(&mut self, id: WidgetID, event: Event) {
        let Some(mapper) = self.mappers.get(&id) else {
            return
        };

        let mapper = mapper.downcast_ref::<Box<dyn Fn(Event) -> Option<Message>>>().expect("invalid mapper");

        self.queue.extend((*mapper)(event).into_iter());
    }
}

#[cfg(test)]
mod test {
    use crate::widget::{Widget, WidgetID};

    use super::WidgetContext;

    struct SimpleWidget(WidgetID);
    impl Widget for SimpleWidget {
        type Event = usize;

        fn id(&self) -> crate::widget::WidgetID { self.0 }
        fn layout_node(&self) -> &dyn crate::layout::LayoutNodeAccess {
            todo!()
        }
        fn poll(&self, _input_info: (), comp: &mut dyn super::WidgetContextAccess<Self::Event>) {
            comp.sink(self.id(), 42);
        }
        fn render(&self, _surface: ()) {
            todo!()
        }
    }

    #[test]
    fn sink_test() {
        let mut wcon = WidgetContext::<()>::default();

        let id1 = WidgetID::new();
        let sw1 = SimpleWidget(id1);

        let id2 = WidgetID::new();
        let sw2 = SimpleWidget(id2);

        wcon.register::<SimpleWidget>(id2, |v| {
            println!("cb: {v}");
            Some(())
        });

        sw1.poll((), &mut wcon);
        sw2.poll((), &mut wcon);

        assert_eq!(wcon.queue, vec![()]);
    }
}
