use crate::{
    component::Component,
    layout::{LayoutNode, LayoutNodeAccess, LayoutNodeContainer},
    platform::{render::{Colour, Painter}, TextInterface},
};

use super::{BuildContext, Widget};

pub struct Frame<C: Component> {
    layout: LayoutNode,
    child: Option<Box<dyn Widget<C>>>,
    _ghost: std::marker::PhantomData<C>,
}

impl<C: Component> Frame<C> {
    pub fn new(bc: &BuildContext) -> Self {
        Self {
            layout: LayoutNode::new(bc.lcache.clone()),
            child: None,
            _ghost: Default::default(),
        }
    }

    pub fn set_child(&mut self, child: Box<dyn Widget<C>>) {
        self.child = Some(child);
        self.layout.relayout();
    }
}

impl<C: Component> LayoutNodeContainer for Frame<C> {
    fn layout_node(&self) -> &LayoutNode {
        &self.layout
    }
    fn layout_child(&self, ndx: usize) -> Option<LayoutNodeAccess> {
        if ndx == 0 { Some(self.child.as_ref()?.layout_node()) } else { None }
    }
    fn layout_child_count(&self) -> usize {
        if self.child.is_some() { 1 } else { 0 }
    }
}

impl<C: Component> Widget<C> for Frame<C> {
    fn poll(&mut self, _input_state: Option<&crate::input::InputState>) -> Vec<C::Msg> {
        vec![]
    }
    fn layout_node(&self) -> LayoutNodeAccess {
        LayoutNodeAccess::new(self)
    }
    fn layout_node_mut(&mut self) -> &mut LayoutNode {
        &mut self.layout
    }
    fn render(&self, painter: &mut dyn Painter, ti: &dyn TextInterface) {
        let rect = self.layout.cached_rect().unwrap();
        painter.fill(rect, Colour::new_rgb(50, 42, 0));
        if let Some(child) = self.child.as_ref() {
            child.render(painter, ti);
        }
    }
}
